
/*********************************************************************
This file is part of the supergetopt library. It contains the definitions
of the supergetopt library functions as well as some useful string
manipulation functions.

Copyright (C) 1997, 1998, 1999 A.P. Russo (apr@planet.net)

The supergetopt library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.
*********************************************************************/


/* Supergetopt: an amazing program to handle your argument needs.
	It verifies options and the format of the arguments to them.
	Handles variable argument lists also.
*/

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include "superget.h"

#define MAXARGS 10		/* no called function can have more than this number of args */
#define MAXOPTS 25		/* only this many options total to superGetOpt() */
#define MAXSTRING 120	/* max of any string passed through */
#define MAXVARARGS 100	/* max for arbitrary args */
#define WILDCARDOPTION "-ExtraArgsOptionXXX"

#define ARGS0 
#define ARGS1 &(optionlist[i].argval[0])
#define ARGS2 ARGS1, &(optionlist[i].argval[1])
#define ARGS3 ARGS2, &(optionlist[i].argval[2])
#define ARGS4 ARGS3, &(optionlist[i].argval[3])
#define ARGS5 ARGS4, &(optionlist[i].argval[4])
#define ARGS6 ARGS5, &(optionlist[i].argval[5])
#define ARGS7 ARGS6, &(optionlist[i].argval[6])
#define ARGS8 ARGS7, &(optionlist[i].argval[7])
#define ARGS9 ARGS8, &(optionlist[i].argval[8])
#define ARGS10 ARGS9, &(optionlist[i].argval[9])

enum 
{
	CHAR,
	SHORT,
	INT,
	FLOAT,
	DOUBLE,
	STRING 
};

typedef union
{
	char c;
	short s;
	int i;
	float f;
	double d;
	char string[MAXSTRING];
} ANYTYPE;

struct optionlist_s 
{
	char name[MAXSTRING];
	int numargs;
	int varflag;
	int argtype[MAXARGS];
	ANYTYPE argval[MAXARGS];
	int (*function)();

	/* these are arrays of size MAXVARARGS, malloced if necessary */
	char *c;
	short *s;
	int *i;
	float *f;
	double *d;
	char **string;
};


static ANYTYPE getval(char *s, int type, int *flag);
static char myread_char(char *s, int *flag);
static unsigned char **local_ucmatrix(int nrl, int nrh, int ncl, int nch);
static short myread_short(char *s, int *flag);
static int myread_int(char *s, int *flag);
static float myread_float(char *s, int *flag);
static double myread_double(char *s, int *flag);
static void local_free_ucmatrix(char **m, int nrl, int nrh, int ncl, int nch);
static int parse_string(char *s, struct optionlist_s *option, int *wildcardflag);
static int malloc_varargs(struct optionlist_s *option);
static int check_if_option(char *s, struct optionlist_s *optionlist, int numopts);
static int free_varargs(struct optionlist_s *option);
static int parse_format(char *s, int *argtypes);


int supergetopt( int argc, char **argv, ... )
{
	va_list ap;
	char *optstring[MAXOPTS+1];
	int optnum, argsleft;
	register int i,j,q;
	int n, x;
	int good;
	int extrargc;
	char **extrargv;
	int found;
	int localwildcardflag, wildcardflag, wildcardnum;
	char wildcardstring[MAXSTRING];
	struct optionlist_s optionlist[MAXOPTS+1];
	int return_val;

	va_start( ap, argv );

	optnum = 0;

	while( (optstring[optnum] = (char *) va_arg(ap, char *)) != (char *)0 )
	{ 
		optionlist[optnum++].function = ( int (*)() ) va_arg(ap, int *);
		if( optnum > MAXOPTS )
		{
			fprintf(stderr, "Too many options in string. more than %d\n",MAXOPTS);
		}
	}

	va_end( ap );

	for( i = 0, wildcardflag = 0 ; i < optnum ; i++ )
	{
		n = optionlist[i].numargs = parse_string(optstring[i],&optionlist[i], &localwildcardflag);
		if( optionlist[i].varflag == 1 )
		{
			malloc_varargs(&optionlist[i]);
		}
		if( localwildcardflag == 1 )
		{
			sprintf(wildcardstring,  "%s%s",WILDCARDOPTION,optstring[i]);

			if( wildcardflag == 1 )
				fprintf(stderr, "Supergetopt Warning: Multiple wildcards found - replacing <%s> with <%s>\n",optstring[wildcardnum],optstring[i]); 

			wildcardflag = 1;	/* to see if ever had a wildcard */
			wildcardnum = i;
		}	
	}

	argsleft = argc;

#ifndef DONTSKIPARG0
	argsleft--;
	argv++;
#endif

	extrargc = 2;
	extrargv = (char **) local_ucmatrix(0,argc+2,0,MAXSTRING);
	if( extrargv == NULL ) return(-999);

	if( argsleft <= 0 )		/* no args to bother with */
	{
		local_free_ucmatrix(extrargv,0,argc+2,0,MAXSTRING);
		for( i = 0, wildcardflag = 0 ; i < optnum ; i++ )
		{
			if( optionlist[i].varflag == 1 )
			{
				free_varargs(&optionlist[i]);
			}
		}
		return(0);
	}

	while( argsleft > 0 )
	{

	for( i = 0 ; i < optnum && argsleft > 0 ; i++ )
	{
		found = 0;
		if( strcmp( argv[0], optionlist[i].name ) == 0 )
		{
			found = 1;
			argsleft--;			
			if( argsleft > 0 ) argv++;
			for( j = 0 ; (j < optionlist[i].numargs && optionlist[i].varflag != 1 && argsleft > 0 ) || (optionlist[i].varflag == 1 && argsleft > 0) ; j++, argsleft--, argv++ )
			{
/*
				if( argsleft <= 0 )
				{
					if( optionlist[i].varflag == 0 )
					{
						fprintf(stderr,"User did not supply enough arguments to option name <%s>\n",optionlist[i].name);
						return( -i-1 );
					}
						
				}
*/
				if( optionlist[i].varflag != 1 )
				{
					switch( optionlist[i].argtype[j] )
					{
						case CHAR: optionlist[i].argval[j] = getval(argv[0], optionlist[i].argtype[j], &good); break;
						case SHORT: optionlist[i].argval[j] = getval(argv[0], optionlist[i].argtype[j], &good); break;
						case INT: optionlist[i].argval[j] = getval(argv[0], optionlist[i].argtype[j], &good); break;
						case FLOAT: optionlist[i].argval[j] = getval(argv[0], optionlist[i].argtype[j], &good); break;
						case DOUBLE: optionlist[i].argval[j] = getval(argv[0], optionlist[i].argtype[j], &good); break;
						case STRING: optionlist[i].argval[j] = getval(argv[0], optionlist[i].argtype[j], &good); 
/* 							if( check_if_option(argv[0], optionlist, optnum) >= 0 && optionlist[i].varflag == 1 ) */
/* 								good = -1; */
							break;
						default: fprintf(stderr, "Bad argtype %d\n",optionlist[i].argtype[j]); return( -i-1 );
					}
				}
				else		/* var arg list */
				{
					optionlist[i].argtype[j] = optionlist[i].argtype[0];
					switch( optionlist[i].argtype[0] )
					{
						case CHAR: optionlist[i].c[j] = myread_char(argv[0],&good); break;
						case SHORT: optionlist[i].s[j] = myread_short(argv[0],&good); break;
						case INT: optionlist[i].i[j] = myread_int(argv[0],&good); break;
						case FLOAT: optionlist[i].f[j] = myread_float(argv[0],&good); break;
						case DOUBLE: optionlist[i].d[j] = myread_double(argv[0],&good); break;
						case STRING: strcpy(optionlist[i].string[j],argv[0]);
								x = check_if_option(argv[0], optionlist, optnum);
 								if( x >= 0 ) /* end of var list */
									good = -2;
								else	
									good = 0; 
								break;
						default: fprintf(stderr, "Bad varargtype %d\n",optionlist[i].argtype[j]); return( -i-1 );
					}
					
				}
				if( good == -1 && optionlist[i].varflag != 1 )
				{
					x = check_if_option(argv[0], optionlist, optnum);
					if( x < 0 )
					{
						fprintf(stderr,"User did not supply correct arguments to option name <%s>\n",optionlist[i].name);
						return(-i-1);
					}
					else 
					{
						fprintf(stderr,"User did not supply enough arguments to option name <%s>\n",optionlist[i].name);
						return( -i-1 );
					}
				}
				else	/* variable arg list */
				if( good == -1 && optionlist[i].varflag == 1 )	/* bad data type */
				{
					x = check_if_option(argv[0], optionlist, optnum);
					if( x < 0 )
					{
						fprintf(stderr, "Var arg list bad data type for option <%s>\n",optionlist[i].name);
						return( -i-1 );
					}
					else	/* next option detected -- end of var list -- move 1 arg back */
					{
						optionlist[i].numargs = j;
						break;		/* get out of numargs loop and call function */
					}
				}
				else
				if( good == -2 && optionlist[i].varflag == 1 )	/* detected end of var list -- move 1 arg back */
				{
					/* Var arg string list terminates at next option */
					optionlist[i].numargs = j;
					break;		/* get out of numargs loop and call function */
				}
				else
				if( argsleft == 1 && optionlist[i].varflag == 1 )	/* end of command line ends var arg list */
				{
					optionlist[i].numargs = j+1;
				}
			}

			if( j != optionlist[i].numargs && optionlist[i].varflag != 1 )
			{
				fprintf(stderr,"User did not supply enough arguments to option name <%s> Expected %d Got %d\n",optionlist[i].name,optionlist[i].numargs,j);
				return( -i-1 );
			}
			if( optionlist[i].function != NULL )	/* got all args for this option */
			{
				if( optionlist[i].varflag == 1 )
				{
					if( optionlist[i].numargs <= 0 )
					{
						fprintf(stderr,"User did not supply enough arguments to option name <%s>. Expected at least 1\n",optionlist[i].name);
						return( -i-1 );
					}
						
					switch( optionlist[i].argtype[0] )
					{
						case CHAR: optionlist[i].function( optionlist[i].c, optionlist[i].numargs ); break;
						case SHORT: optionlist[i].function( optionlist[i].s, optionlist[i].numargs ); break;
						case INT: optionlist[i].function( optionlist[i].i, optionlist[i].numargs ); break;
						case FLOAT: optionlist[i].function( optionlist[i].f, optionlist[i].numargs ); break;
						case DOUBLE: optionlist[i].function( optionlist[i].d, optionlist[i].numargs ); break;
						case STRING: optionlist[i].function( optionlist[i].string, optionlist[i].numargs ); break;
					}
/* 					free_varargs( &optionlist[i] ); */ /* do it at end */
				}
				else
				{
					switch( optionlist[i].numargs )
					{
						case 0: optionlist[i].function(); break;
						case 1: optionlist[i].function( ARGS1 ); break;
						case 2: optionlist[i].function( ARGS2 ); break;
						case 3: optionlist[i].function( ARGS3 ); break;
						case 4: optionlist[i].function( ARGS4 ); break;
						case 5: optionlist[i].function( ARGS5 ); break;
						case 6: optionlist[i].function( ARGS6 ); break;
						case 7: optionlist[i].function( ARGS7 ); break;
						case 8: optionlist[i].function( ARGS8 ); break;
						case 9: optionlist[i].function( ARGS9 ); break;
						case MAXARGS: optionlist[i].function( ARGS10 ); break;
						default: fprintf(stderr, "CALL_FUNCTION: Bad number of arguments %d\n", optionlist[i].numargs);
							return(-1);
					}
				}

			}
			
			i = -1;  /* experimental bug fix? */
		}
		else
			found = 0;
	}
	if( found == 0 )
	{
		/* print_option( &optionlist[i] ); */ 
		extrargc++;
/* 		extrargv[extrargc-1] = argv[0]; */
		strcpy(extrargv[extrargc-1], argv[0]);
		if( argsleft > 0 ) argv++;
		argsleft--;
	}

	}

	return_val = 0;
	if( extrargc > 2 )
	{
		if( wildcardflag == 1 )
		{
			/* handle extra arg wildcard by making extraargv look like command line with
				WILDCARDOPTION flag. Then call supergetopt to parse it! */

			strcpy(extrargv[0], "supergetopt");
			strcpy(extrargv[1], WILDCARDOPTION);

			q = supergetopt(extrargc,extrargv, wildcardstring, optionlist[wildcardnum].function, (char *) 0 );
			if( q != 0 )
			{
				fprintf(stderr,"Problem with flagless arguments. Returning %d\n",q-1);
				return_val = (  q-1 );
			}
		}
		else
		{
			for( q = 2 ; q < extrargc ; q++ )
			{
				fprintf(stderr,"Extra arg #%d is <%s>\n",q,extrargv[q]);
			}
			/* *origargv = extrargv[2]; */
			return_val = ( extrargc-2 );
		}

	}
	
	for( i = 0, wildcardflag = 0 ; i < optnum ; i++ )
	{
		if( optionlist[i].varflag == 1 )
		{
			free_varargs(&optionlist[i]);
		}
	}

	local_free_ucmatrix(extrargv, 0, argc+2, 0, MAXSTRING);
	
	return( return_val );
}


static int parse_string(char *s, struct optionlist_s *option, int *wildcardflag)
{

	char *scopy;
	int len, n, m , z, offset;

	len = strlen( s );

	if( len <= 0 )
		return(0);

	scopy = (char *) strdup( s );

	if( len < 2 )
	{
		fprintf(stderr, "Parse_string: option has zero length <%s>\n", s);
		free(scopy); /* memleak!!!! found by purify!!! */
		return(-1);
	}

	n = strgrep( scopy, ":" );	/* explicit list implied */
	m = strgrep( scopy, ":*" );	/* variable argument list implied */

	free(scopy); /* memleak!!!! found by purify!!! */

	if( m == n && n >= 0 )		/* vararg list implied */
	{
		offset = 2;
		option->varflag = 1;
	}
	else			/* explicit list implied */
	{
		offset = 1;
		option->varflag = 0;
	}
	
	if( n >= 0 )
	{
		strncpy(option->name, s, n);
		option->name[n] = '\0';

		if( n == 0 )		/* no name ==> wildcard */
			*wildcardflag = 1;
		else
			*wildcardflag = 0;

		if( n >= len - 1 )
		{
			fprintf(stderr, "Parse_String: No formats given in <%s>\n",s /* was scopy */);
			return(0);
		}
		else
		{
			if( option->varflag == 0 )
				return( parse_format( &s[n+offset], option->argtype ) );
			else
			{
				if( (z = parse_format( &s[n+offset], option->argtype )) == 1 )
					return(0);  /* newly adden 9-27-92. bug if gave vararg option but no list */
				else
				{
					fprintf(stderr,"Not allowed to mix types in an option with a var arg list\n");
					return(-1234);
				}
			}
		}
	}
	else
	{
		*wildcardflag = 0;
		strcpy( option->name, s );
		return( 0 );
	}

}

static int parse_format(char *s, int *argtypes)
{
	static char *scopy, string[MAXSTRING];
	int len;
	int i, numargs;
	static char *arg[MAXARGS];

	while ( s[0] == ' ' )
		s++;
		
	len = strlen( s );
	scopy = (char *) strdup( s );



	for( numargs = -1, arg[++numargs] = strtok(scopy, "%");
		(arg[++numargs] = strtok( (char *) NULL, "%")) != NULL && numargs < MAXARGS-1; )
		;

	if( strgrep(s,"%") != 0 )
	{
		fprintf(stderr, "Bad format in <%s> near <%s>\n", s, arg[0]);
	}

	for( i = 0 ; i < numargs ; i++ )
	{
		strcpy(string,"%");
		strcat(string, arg[i]);

		if( strgrep(string, "%f") != -1 )
			argtypes[i] = (int) FLOAT;
		else
		if( strgrep(string, "%lf") != -1 )
			argtypes[i] = (int) DOUBLE;
		else
		if( strgrep(string, "%c") != -1 )
			argtypes[i] = (int) CHAR;
		else
		if( strgrep(string, "%hd") != -1 )
			argtypes[i] = (int) SHORT;
		else
		if( strgrep(string, "%d") != -1 )
			argtypes[i] = (int) INT;
		else
		if( strgrep(string, "%s") != -1 )
			argtypes[i] = (int) STRING;
		else
		{
			fprintf(stderr, "Parse_Format: Bad format <%s> len=%d arg0=%s\n",string,len,arg[0]);
#if 0
			free(scopy); /* memleak!!!! found by purify!!! */
#endif
			return(-1);
		}
	}
	
#if 1
	free(scopy); /* memleak!!!! found by purify!!! */
#endif

	return( numargs );
}


static ANYTYPE getval(char *s, int type, int *flag)
{
	ANYTYPE value;

	*flag = 0;

	switch( type )
	{
		case CHAR:
			if( sscanf( s, "%c", &value.c ) != 1 )
			{
				fprintf(stderr," Getval: Bad argument. Expected character\n");
				*flag = -1;
				return( value );
			}
			else  return( value ); 	
		case SHORT:
			if( sscanf( s, "%hd", &value.s ) != 1 )
			{
				fprintf(stderr," Getval: Bad argument. Expected short\n");
				*flag = -1;
				return( value );
			}
			else return( value ); 	
		case INT:
			if( sscanf( s, "%d", &value.i ) != 1 )
			{
				fprintf(stderr," Getval: Bad argument. Expected integer\n");
				*flag = -1;
				return( value );
			}
			else return( value ); 	
		case FLOAT:
			if( sscanf( s, "%f", &value.f ) != 1 )
			{
				fprintf(stderr," Getval: Bad argument. Expected float\n");
				*flag = -1;
				return( value );
			}
			else return( value ); 	
		case DOUBLE:
			if( sscanf( s, "%lf", &value.d ) != 1 )
			{
				fprintf(stderr," Getval: Bad argument. Expected double\n");
				*flag = -1;
				return( value );
			}
			else return( value ); 	
		case STRING:
			strcpy( value.string, s );
			return( value );
		default:
			fprintf(stderr, "Getval: Bad argument type %d\n",type);
			*flag = -1;
			return( value );
	} 	
}

static char myread_char(char *s, int *flag)
{
	char x;
	*flag = 0;
	if( sscanf( s, "%c", &x ) != 1 || strlen(s) > 1 )
	{
		*flag = -1;
	}
	return( x );
} 	
static short myread_short(char *s, int *flag)
{
	short x;
	*flag = 0;
	if( sscanf( s, "%hd", &x ) != 1 )
	{
		*flag = -1;
	}
	return( x );
} 	
static int myread_int(char *s, int *flag)
{
	int x;
	*flag = 0;
	if( sscanf( s, "%d", &x ) != 1 )
	{
		*flag = -1;
	}
	return( x );
} 	
static float myread_float(char *s, int *flag)
{
	float x;
	*flag = 0;
	if( sscanf( s, "%f", &x ) != 1 )
	{
		*flag = -1;
	}
	return( x );
} 	
static double myread_double(char *s, int *flag)
{
	double x;
	*flag = 0;
	if( sscanf( s, "%lf", &x ) != 1 )
	{
		*flag = -1;
	}
	return( x );
} 	
 
static int check_if_option(char *s, struct optionlist_s *optionlist, int numopts)
{

	register int i;

	for( i = 0 ; i < numopts ; i++ )
	{
		if( strcmp( s, optionlist[i].name ) == 0 )
			return( i );
	}

	return( -1 );
}
	
static int malloc_varargs(struct optionlist_s *option)
{
/* 	fprintf(stderr,"malloc_varargs called\n"); */

	switch( option->argtype[0] )
	{
		case CHAR: option->c = (char *) malloc( sizeof(char) * MAXVARARGS );
				if( option->c == NULL ) { fprintf(stderr,"Bad malloc in malloc_varargs\n"); return(-1); }
				break;
		case SHORT: option->s = (short *) malloc( sizeof(short) * MAXVARARGS );
				if( option->s == NULL ) { fprintf(stderr,"Bad malloc in malloc_varargs\n"); return(-1); }
				break;
		case INT: option->i = (int *) malloc( sizeof(int) * MAXVARARGS );
				if( option->i == NULL ) { fprintf(stderr,"Bad malloc in malloc_varargs\n"); return(-1); }
				break;
		case FLOAT: option->f = (float *) malloc( sizeof(float) * MAXVARARGS );
				if( option->f == NULL ) { fprintf(stderr,"Bad malloc in malloc_varargs\n"); return(-1); }
				break;
		case DOUBLE: option->d = (double *) malloc( sizeof(double) * MAXVARARGS );
				if( option->d == NULL ) { fprintf(stderr,"Bad malloc in malloc_varargs\n"); return(-1); }
				break;
		case STRING: option->string = (char **) local_ucmatrix(0,MAXVARARGS,0,MAXSTRING);
				if( option->string == NULL ) { fprintf(stderr,"Bad malloc in malloc_varargs\n"); return(-1); }
				break;
	}

	return( 0 );

}

static int free_varargs(struct optionlist_s *option)
{

	if( option->varflag == 1 )
	{
		switch( option->argtype[0] )
		{
			case CHAR: free(option->c); break;
			case SHORT: free(option->s); break;
			case INT: free(option->i); break;
			case FLOAT: free(option->f); break;
			case DOUBLE: free(option->d); break;
			case STRING:	
					local_free_ucmatrix(option->string, 0,MAXVARARGS,0,MAXSTRING); 
					break;
		}
	}

	return( 0 );
}


static unsigned char **local_ucmatrix(int nrl, int nrh, int ncl, int nch)
{
	int i;
	unsigned char **m;
	m=( unsigned char **)malloc( (nrh-nrl+1)*sizeof(unsigned  char*));
	if (!m) {
		fprintf(stderr,"allocation failure 1 in imatrix()");
		return(NULL);
	}
	m -= nrl;

	for(i=nrl;i<=nrh;i++) {
		m[i]=( unsigned char *)malloc( (nch-ncl+1)*sizeof( unsigned char));
		if (!m[i]) {
			fprintf(stderr,"allocation failure 2 in imatrix(): nrh=%d nch=%d",nrh,nch);
			return(NULL);
		}
		m[i] -= ncl;
	}

/* 	fprintf(stderr,"local_ucmatrix: at 0x%x grabbed %d x %d = %d bytes\n",m,nrh,nch,nrh*nch); */
	return m;
}

static void local_free_ucmatrix(char **m, int nrl, int nrh, int ncl, int nch)
{
	int i;

	for(i=nrh;i>=nrl;i--) 
	{
		if( m[i]+ncl != NULL )
			free((unsigned char*) (m[i]+ncl));
	}
	free((unsigned char*) (m+nrl));
/* 	fprintf(stderr,"local_free_ucmatrix: at 0x%x freed %d x %d = %d bytes nrl=%d\n",m,nrh,nch,nrh*nch,nrl); */
}

